/* Functions to compute the likelihood for a given species tree */

#include <stdio.h>
#include <math.h>
#include "main.h"
#include "qsort.h"

void double_qsort(double *arr, unsigned n) {
#define double_lt(a,b) ((*a)<(*b))
QSORT(double, arr, n, double_lt);
}

int FindParent(int target) {

  int i, j, parent=0;

  for (j=0; j<2; j++) {
    for (i=0; i<nspecies; i++) {
      if (ppTwoRowS[j][i]==target) {
        parent = i+nspecies+1;
        break;
      }
    }
  }

  /*  if (parent==0) { printf("Could not find parent for node %d\n",target); }*/
  return(parent);

}


int FindParentTree(int target, int tree) {

  int i, j, parent=0;

  for (j=0; j<2; j++) {
    for (i=0; i<ntaxa; i++) {
      if (pppTR[tree][j][i]==target) {
        parent = i+ntaxa+1;
        break;
      }
    }
  }

  /*  if (parent==0) { printf("Could not find parent for node %d\n",target); }*/
  return(parent);

}



/* This creates a vector that tells us which lineages are available */
/* to coalesce on a particular species tree branch                  */

int AvailCoal(int branch) {

  int ii, current=0, counter=0;

  for (ii=1; ii<nspecies+1; ii++) {
    current = ii;
    while (current != nspecies+1) {
      current = FindParent(current);
      if (current == branch) {
	CoalVec[counter] = ii;
	counter+=1;
	break;
      }
    }

  }

  /* for (ii=0; ii<2*ntaxa; ii++) printf("%d %d\n",ii,CoalVec[ii]);
     printf("counter is %d\n",counter);*/

  return counter;
}


double FindMRCA(int tax1, int tax2, int tree_num) {

  int parent1, parent2, check=0;
  double time_mrca;

  parent1 = FindParentTree(tax1,tree_num);
  parent2 = FindParentTree(tax2,tree_num);

  while (parent1 != ntaxa+1 && check != 1) {
    while (parent2 != ntaxa+1 && check != 1) {
      if (parent1 == parent2) {
	check = 1;
	break;
      }
      parent2 = FindParentTree(parent2,tree_num);
    }
    if (check == 1) break;
    else {
      parent1 = FindParentTree(parent1,tree_num);
      parent2 = FindParentTree(tax2,tree_num);
    }
  }
  time_mrca = ppTimeVec[tree_num][parent1];
  /*if ((parent1 == ntaxa+1 || parent2 == ntaxa+1) && check == 0) {
    time_mrca = ppTimeVec[tree_num][ntaxa+1];
    printf("mrca is ntaxa+1\n\n");
    check = 1;
    }*/

  return time_mrca;
}


void GetSpecCon() {

  int i, j, k, l, m;
  double time, min_time;

  for (i=1; i<nspecies+1; i++) {
    
    for (j=i+1; j<nspecies+1; j++) {

      min_time = 1000.0;
      for (k=0; k<SpecCount[i]; k++) {
	for (l=0; l<SpecCount[j]; l++) {
	  for (m=0; m<ngtrees; m++){
	  time = FindMRCA(SpecVec[i][k],SpecVec[j][l],m);
	  if (time < min_time && time>=0.0) min_time = time;
	  }
	}
      }
      ppSpecCon[i][j] = ppSpecCon[j][i] = min_time;

    }
  }

  /* Check */

  /*printf("\n ppSpecCon is:\n");*/
  printf("\nThe D_AB matrix is:\n");
    for  (i=1; i<nspecies+1; i++) {
      for (j=1; j<nspecies+1; j++) {
	printf("%f ",ppSpecCon[i][j]);
	fflush(0);
      }
      printf("\n");
    }
  
  

}







double GetBranchLik(double *CoalTimes, int ncoal, int navail, int node) {

  int nn;
  double blik = 0.0;


  //  printf("node is %d, ncoal is %d, navail is %d\n",node,ncoal,navail);

  /* printf("before qsort, times are:\n");
  for (mm=0; mm<ncoal; mm++) printf("%f ",CoalTimes[mm]);
  printf("\n");*/
  double_qsort(CoalTimes,ncoal);
  /*printf("after qsort, times are:\n");
  for (mm=0; mm<ncoal; mm++) printf("%f ",CoalTimes[mm]);
  printf("\n");*/
  //printf("          In GBL, ncoal is %d, navail is %d, node is %d, TV parent is %f, parent is %f\n",ncoal,navail,node,TimeVecS[FindParent(node)],TimeVecS[node]);

  if (ncoal==0) {
    blik = exp(-1.0*navail*(navail-1)*(TimeVecS[FindParent(node)]-TimeVecS[node]));
  }

  else {
    blik = (2.0)*exp(-1.0*navail*(navail-1)*(CoalTimes[0]-TimeVecS[node]));
    navail -= 1;
    for (nn=1; nn<ncoal; nn++) {
      blik = blik * (2.0)*exp(-1.0*navail*(navail-1)*(CoalTimes[nn]-CoalTimes[nn-1]));
      navail-=1;
    }
    if (navail>1) blik = blik * exp(-1.0*navail*(navail-1)*(TimeVecS[FindParent(node)] - CoalTimes[ncoal-1]));
  }

  //  printf("for branch %d blik is %f\n",node,blik);
  return log(blik);

}


void CalcLik(int ntree) {
  
  int aa, i, ii, j, jj, kk, ks, mm, nn, pp, num_adjusted=0, num_coal=0, num_coal2=0, num_coal3=0, num_coal4=0, num_lin_coal=0, found_it=0;
  double ppCoalMat[ntaxa][ntaxa], pCoalTimes[ntaxa], pPriorCoalTimes[ntaxa], pFutureCoalTimes[ntaxa], coal_time;

  int tempcount=0, mycount=0;

  /********************************************/
  /* first do calcs for all external branches */
  /********************************************/

  for (ii=1; ii<nspecies+1; ii++) {

    num_coal2 = num_coal3 = 0;
    for (jj=0; jj<ntaxa; jj++) pCoalTimes[jj] = pPriorCoalTimes[jj] = 0.0;
    num_lin_coal = SpecCount[ii]; 
    for (jj=0; jj<num_lin_coal; jj++) { //Check to see if taxa are in this tree; add to list if yes.
      found_it=0;
      for (aa=0; aa<ntaxa-1; aa++) {
	//printf("ii is %d, jj is %d, aa is %d, pppTR[ntree][0][aa] is %d, pppTR[ntree][1][aa] is %d\n",ii,jj,aa,pppTR[ntree][0][aa],pppTR[ntree][1][aa]);
	if ((pppTR[ntree][0][aa] == SpecVec[ii][jj]) || (pppTR[ntree][1][aa] == SpecVec[ii][jj])) {
	  CoalVec2[jj] = SpecVec[ii][jj];
	  found_it=1;
	}
      }
      //printf("%d %d \n",jj,CoalVec2[jj]);
    }
    if (found_it != 1) num_lin_coal--;

    for (kk=0; kk<num_lin_coal; kk++) {
      for (jj=kk+1; jj<num_lin_coal; jj++) {
	coal_time = FindMRCA(CoalVec2[kk],CoalVec2[jj],ntree);
	if (coal_time < TimeVecS[FindParent(ii)]) {
	  mm = 0;
	  while (coal_time != pCoalTimes[mm] && mm < num_coal2) mm++;
	  if (mm == num_coal2) {
	    num_coal2+=1;
	    pCoalTimes[mm] = coal_time;
	  }	      
	}
	
      }
    }
    
    /*printf("For %d, number of coalescent events is %d, number of available lineages is %d\n",ii,num_coal2,num_lin_coal);*/
	ppTipMat[ntree][ii] = GetBranchLik(pCoalTimes,num_coal2,num_lin_coal,ii);

  }

  /***********************************/
  /* now do calcs for internal nodes */
  /***********************************/

  for (ii=nspecies+1; ii<2*nspecies; ii++) recalced[ii]=0;
  for (ii=0; ii<nspecies+1; ii++) recalced[ii]=1;
  
  while (num_adjusted<nspecies-2) {
    
    for (ii=0; ii<nspecies-1; ii++) {
      
      if (recalced[ppTwoRowS[0][ii]]==1 && recalced[ppTwoRowS[1][ii]]==1 && recalced[ii+nspecies+1]==0) {
	
	/***********************************/
	/* do calcs for node ii+nspecies+1 */
	/***********************************/
	num_coal2 = num_coal3 = 0;
	tempcount=mycount=0;
	for (jj=0; jj<2*nspecies+1; jj++) CoalVec[jj] = 0;
	for (jj=0; jj<ntaxa; jj++) pCoalTimes[jj] = pPriorCoalTimes[jj] = 0.0;
	num_coal = AvailCoal(ii+nspecies+1);
	num_lin_coal = 0;
	for (i=0; i<num_coal; i++) {
	  for (ks=0; ks<SpecCount[CoalVec[i]]; ks++) {
	    found_it=0;
	    for (aa=0; aa<ntaxa-1; aa++) {
	      if ((pppTR[ntree][0][aa] == SpecVec[CoalVec[i]][ks]) || (pppTR[ntree][1][aa] == SpecVec[CoalVec[i]][ks])) {
		CoalVec2[num_lin_coal] = SpecVec[CoalVec[i]][ks];
		found_it = 1;
	      }
	    }
	    //printf("for ks=%d,CoalVec2[num_lin_coal] is %d\n",ks,CoalVec2[num_lin_coal]);
	    if (found_it == 1) num_lin_coal++;
	  }
	}
	//printf("num_lin_coal is %d\n",num_lin_coal);

	for (kk=0; kk<num_lin_coal; kk++) {
	  for (jj=kk+1; jj<num_lin_coal; jj++) {
	    if (CoalVec2[jj] != 0) coal_time = FindMRCA(CoalVec2[kk],CoalVec2[jj],ntree);
	    if (coal_time >= TimeVecS[ii+nspecies+1] && coal_time < TimeVecS[FindParent(ii+nspecies+1)]) {
	      mm = 0;
	      while (coal_time != pCoalTimes[mm] && mm < num_coal2) mm++;
	      if (mm == num_coal2) {
		num_coal2+=1;
		pCoalTimes[mm] = coal_time;
	      }	      
	    }
	    if (coal_time < TimeVecS[ii+nspecies+1]) {
	      tempcount=tempcount+1;
	      nn=0;
	      while (coal_time != pPriorCoalTimes[nn] && nn < num_coal3) nn++;
	      if (nn == num_coal3) {
		num_coal3+=1;
		pPriorCoalTimes[nn] = coal_time;
	      }
	    }
	  }
	  if (tempcount>0) mycount+=1;
	  tempcount = 0;
	}
	

	/*printf("For %d, number of coalescent events is %d, number of available lineages is %d\n",ii+nspecies+1,num_coal2,num_lin_coal-num_coal3);*/
	ppLikMat[ntree][ii] = GetBranchLik(pCoalTimes,num_coal2,num_lin_coal-mycount,ii+nspecies+1);

	/***********************/
	/* done for ii+nspecies+1 */
	/***********************/
	
	recalced[ii+nspecies+1] = 1;
	num_adjusted++;
	
      }
      
    }
    
  }
  
  /***************/
  /* now do root */
  /***************/
  
  ii=0;
  num_coal2 = num_coal3 = 0;
  for (jj=0; jj<2*nspecies+1; jj++) CoalVec[jj] = 0;
  for (jj=0; jj<nspecies; jj++) pCoalTimes[jj] = pPriorCoalTimes[jj] = 0.0;
  num_coal = AvailCoal(ii+nspecies+1);
  num_lin_coal = 0;
  for (i=0; i<num_coal; i++) {
    for (ks=0; ks<SpecCount[CoalVec[i]]; ks++) {
      found_it=0;
      for (aa=0; aa<ntaxa-1; aa++) {
	if ((pppTR[ntree][0][aa] == SpecVec[CoalVec[i]][ks]) || pppTR[ntree][1][aa] == SpecVec[CoalVec[i]][ks]) {
	  CoalVec2[num_lin_coal] = SpecVec[CoalVec[i]][ks];
	  found_it=1;
	}
      }
      if (found_it==1) num_lin_coal++;
    }
  }

  for (kk=0; kk<num_lin_coal; kk++) {
    for (jj=kk+1; jj<num_lin_coal; jj++) {
      coal_time = FindMRCA(CoalVec2[kk],CoalVec2[jj],ntree);
      if (coal_time >= TimeVecS[ii+nspecies+1]) {
	mm = 0;
	while (coal_time != pCoalTimes[mm] && mm < num_coal2) mm++;
	if (mm == num_coal2) {
	  num_coal2+=1;
	  pCoalTimes[mm] = coal_time;
	}	
      }

      if (coal_time < TimeVecS[ii+nspecies+1]) {
	nn=0;
	while (coal_time != pPriorCoalTimes[nn] && nn < num_coal3) nn++;
	if (nn == num_coal3) {
	  num_coal3+=1;
	  pPriorCoalTimes[nn] = coal_time;
	}
      }
    }
  }

  //printf("\t\tnumber of coalescent events is %d, number of prior coalescences is %d, Number of lineages is %d, number of available lineages is %d\n",num_coal2,num_coal3,num_lin_coal,num_lin_coal-num_coal3);
  ppLikMat[ntree][0] = GetBranchLik(pCoalTimes,num_coal2,num_lin_coal-num_coal3,ii+nspecies+1);  //Second num_coal2 used to num_lin_coal-num_coal3
  /*printf("likelihood is %f\n\n",ppLikMat[ntree][0]);*/

}


/* For any tree, can find ML branch lengths by starting at root and working down */

void GetMLLengths() {

  int ii, jj, kk, num_adjusted=0, num_coalL=0, num_coalR=0;
  int current_node=0, find_next_node=0;
  double min_time;

  for (ii=nspecies+1; ii<2*nspecies; ii++) recalced[ii]=0;
  for (ii=0; ii<nspecies+1; ii++) recalced[ii]=1;

  /* Find time for root*/
  current_node = nspecies+1;

  for (jj=0; jj<2*nspecies+1; jj++) {
    CoalVecL[jj] = 0;
    CoalVecR[jj] = 0;
  }

  if (ppTwoRowS[0][0]>nspecies) {

    num_coalL = AvailCoal(ppTwoRowS[0][0]);
    for (jj=0; jj<2*nspecies+1; jj++) CoalVecL[jj] = CoalVec[jj];
  }
  else {
    num_coalL = 1;
    CoalVecL[0] = ppTwoRowS[0][0];
  }

  if (ppTwoRowS[1][0]>nspecies) {
    num_coalR = AvailCoal(ppTwoRowS[1][0]);
    for (jj=0; jj<2*nspecies+1; jj++) CoalVecR[jj] = CoalVec[jj];
  }
  else {
    num_coalR = 1;
    CoalVecR[0] = ppTwoRowS[1][0];
  }

  min_time = 1000.0;
  for (kk=0; kk<num_coalL; kk++) {
    for (jj=0; jj<num_coalR; jj++) if (ppSpecCon[CoalVecL[kk]][CoalVecR[jj]] < min_time) min_time = ppSpecCon[CoalVecL[kk]][CoalVecR[jj]];
  }
  if (min_time < 1000.0) TimeVecS[nspecies+1] = min_time;

  recalced[current_node] = 1;
  if (ppTwoRowS[0][current_node-(nspecies+1)]>nspecies+1) current_node = ppTwoRowS[0][current_node-(nspecies+1)];
  else current_node = ppTwoRowS[1][current_node-(nspecies+1)];

  
  while (num_adjusted<nspecies-2) {

    /***********************************/
    /* do calcs for node ii+nspecies+1 */
    /***********************************/
    for (jj=0; jj<2*nspecies+1; jj++) {
      CoalVecL[jj] = 0;
      CoalVecR[jj] = 0;
    }
    if (ppTwoRowS[0][current_node-(nspecies+1)]>nspecies) {
      num_coalL = AvailCoal(ppTwoRowS[0][current_node-(nspecies+1)]);
      for (jj=0; jj<2*nspecies+1; jj++) CoalVecL[jj] = CoalVec[jj];
    }
    else {
      num_coalL = 1;
      CoalVecL[0] = ppTwoRowS[0][current_node-(nspecies+1)];
    }

    if (ppTwoRowS[1][current_node-(nspecies+1)]>nspecies) {
      num_coalR = AvailCoal(ppTwoRowS[1][current_node-(nspecies+1)]);
      for (jj=0; jj<2*nspecies+1; jj++) CoalVecR[jj] = CoalVec[jj];
    }
    else {
      num_coalR = 1;
      CoalVecR[0] = ppTwoRowS[1][current_node-(nspecies+1)];
    }
    
    min_time = 1000.0;
    for (kk=0; kk<num_coalL; kk++) {
      for (jj=0; jj<num_coalR; jj++) if ((ppSpecCon[CoalVecL[kk]][CoalVecR[jj]] < min_time) && (ppSpecCon[CoalVecL[kk]][CoalVecR[jj]] <= TimeVecS[FindParent(current_node)])) {
	min_time = ppSpecCon[CoalVecL[kk]][CoalVecR[jj]];
      }
    }
    if (min_time < 1000.0) TimeVecS[current_node] = min_time;
    else TimeVecS[current_node] = TimeVecS[FindParent(current_node)];
    
    /***********************/
    /* done for ii+nspecies+1 */
    /***********************/
    recalced[current_node] = 1;
    num_adjusted++;

    /* find next internal node to be modified */
    
    if (ppTwoRowS[0][current_node-(nspecies+1)]>nspecies+1 && recalced[ppTwoRowS[0][current_node-(nspecies+1)]] == 0) current_node = ppTwoRowS[0][current_node-(nspecies+1)];
    else if (ppTwoRowS[1][current_node-(nspecies+1)]>nspecies+1 && recalced[ppTwoRowS[1][current_node-(nspecies+1)]] == 0) current_node = ppTwoRowS[1][current_node-(nspecies+1)];
    else {
      
      find_next_node = 0;
      while ((current_node != nspecies+1) && (find_next_node == 0)) {

        current_node = FindParent(current_node);
        if (ppTwoRowS[0][current_node-(nspecies+1)]>nspecies+1 && recalced[ppTwoRowS[0][current_node-(nspecies+1)]] == 0) {
          current_node = ppTwoRowS[0][current_node-(nspecies+1)];
          find_next_node = 1;
        }

        else if (ppTwoRowS[1][current_node-(nspecies+1)]>nspecies+1 && recalced[ppTwoRowS[1][current_node-(nspecies+1)]] == 0) {
          current_node = ppTwoRowS[1][current_node-(nspecies+1)];
          find_next_node =1;
          
        }

      }
      
    }
    
  }

}


double GetLik() {

  int ii, jj;
  double tlik=0.0, alik=0.0;

  for (jj=0; jj<ngtrees; jj++) {

    CalcLik(jj);
    tlik = 0.0;
    for (ii=0; ii<nspecies-1; ii++) {
      tlik = tlik + ppLikMat[jj][ii];
    }
    for (ii=1; ii<nspecies+1; ii++) {
      tlik = tlik + ppTipMat[jj][ii];
    }
    alik = alik + tlik;
  }

  return alik;

}








